from typing import Optional, Union, List
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
import itertools
import traceback

import numpy as np
import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_single_batch",
    "update_batch",
    "get_batches",
    "update_batch_values",
    "update_batch_feature_values",
    "update_batch_features_and_values",
    "update_vector_batch_values",
    "create_or_update_batches",
    "get_batch_relations",
    "get_single_batch_relation",
]


# Batches
def get_batch_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the configured batch types from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing batch types
    """

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.batches.get_batch_types(
            page=page_num, page_size=page_size, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_batch_type_details(
    batch_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get details about a single batch type from the OIAnalytics API

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A tuple of 3 DataFrames listing various properties of the batch type:
        - Steps
        - Data
        - Features
    """

    # Query endpoint
    response = endpoints.batches.get_batch_type_details(
        batch_type_id=batch_type_id, api_credentials=api_credentials
    )

    # Split content
    steps = response["steps"]
    data = response["dataList"]
    features = response["features"]

    # Format dataframes
    if len(steps) > 0:
        df_steps = pd.DataFrame(steps).set_index("id")
    else:
        df_steps = pd.DataFrame(columns=["id", "name", "localisationKeys"]).set_index(
            "id"
        )

    if len(data) > 0:
        df_data = pd.DataFrame(data).set_index("id")
    else:
        df_data = pd.DataFrame(
            columns=["id", "dataType", "reference", "description"]
        ).set_index("id")

    if len(features) > 0:
        df_features = pd.DataFrame(features).set_index("id")
    else:
        df_features = pd.DataFrame(columns=["id", "key"]).set_index("id")

    # Output
    return df_steps, df_data, df_features


def get_single_batch(
    batch_type_id: str,
    batch_id: str,
    split_steps_and_values: bool = True,
    extract_from_localisation: Optional[str] = "value",
    expand_localisation: bool = True,
    extract_from_values: Optional[str] = "reference",
    expand_value_ids: bool = True,
    extract_from_features: Optional[str] = "value",
    expand_features: bool = True,
    append_unit_to_description: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> Union[pd.DataFrame, tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Retrieve the detail of a specific batch with its batch-type ID and batch ID.

    Parameters
    ----------
    batch_type_id : str
        The ID of the batch type the batch belongs to
    batch_id : str
        The ID of the batch to get.
    split_steps_and_values : bool, default True
        Whether to split the output into two DataFrames, one containing the steps details and the other containing thz values details.
    extract_from_localisation : str, optional, default 'value'
        Attribute to extract from the column 'localisation'.
    expand_localisation : bool, default True
        Whether to expand the column 'tagValuesLocalisation'.
    extract_from_values : str, default 'reference'
        Attribute to extract from 'values'.
    expand_value_ids : bool, default True
        Whether to expand the column 'values'
    extract_from_features : str, default 'value'
        Attribute to extract from column 'features'.
    expand_features : str, default 'value'
        Whether to expand column 'features'.
    append_unit_to_description : bool, default 'value'
        Whether to append unit to the column 'description'. Used if extract_from_values is 'description'.
    api_credentials : _credentials.OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pandas.DataFrame
        DataFrame or tuple of DataFrame with details of the requested batch.
    """
    # Args validation
    if extract_from_localisation not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_localisation': {extract_from_localisation}"
        )

    if extract_from_localisation is None and expand_localisation is True:
        raise ValueError(
            "Localisation cannot be expanded if 'extract_from_values' is None"
        )

    if extract_from_values not in ["id", "reference", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_values': {extract_from_values}"
        )

    if extract_from_values is None and expand_value_ids is True:
        raise ValueError("Values cannot be expanded if 'extract_from_values' is None")

    if extract_from_features not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_features': {extract_from_features}"
        )

    if extract_from_features is None and expand_features is True:
        raise ValueError(
            "Features cannot be expanded if 'extract_from_features' is None"
        )

    # Query endpoint
    response = endpoints.batches.get_single_batch(
        batch_type_id=batch_type_id, batch_id=batch_id, api_credentials=api_credentials
    )

    # Format response
    df = pd.Series(response).to_frame().T

    # Expected columns if content is empty
    if df.shape[0] == 0:
        df = pd.DataFrame(
            columns=["id", "name", "timestamp", "steps", "values", "features"]
        )

    # Rename columns
    df.rename(
        columns={
            "id": "batch_id",
            "name": "batch_name",
            "timestamp": "batch_timestamp",
        },
        inplace=True,
    )

    # Parse dates
    df["batch_timestamp"] = pd.to_datetime(df["batch_timestamp"], format="ISO8601")

    # Set index
    df.set_index(["batch_id", "batch_name", "batch_timestamp"], inplace=True)

    # Split steps and values
    if split_steps_and_values is True:
        # Split dataframe
        df_steps = df.drop(columns=["values", "features"])
        df_values = df.drop(columns="steps")

        # Format steps
        df_steps = df_steps.explode("steps")
        df_steps = utils.expand_dataframe_column(
            df_steps,
            "steps",
            add_prefix=False,
            expected_keys=[
                "step",
                "start",
                "end",
                "localisationType",
                "assetLocalisation",
                "tagValuesLocalisation",
            ],
        )
        df_steps = utils.expand_dataframe_column(
            df_steps, "step", expected_keys=["id", "name"]
        )

        df_steps["start"] = pd.to_datetime(df_steps["start"], format="ISO8601")
        df_steps["end"] = pd.to_datetime(df_steps["end"], format="ISO8601")

        if extract_from_localisation == "id":
            df_steps["assetLocalisation"] = df_steps["assetLocalisation"].map(
                lambda asset: None if asset is None else asset["id"]
            )

            df_steps["tagValuesLocalisation"] = df_steps["tagValuesLocalisation"].apply(
                lambda full_loc: {loc["tagKey"]["id"]: loc["id"] for loc in full_loc}
            )

        elif extract_from_localisation == "value":
            df_steps["assetLocalisation"] = df_steps["assetLocalisation"].map(
                lambda asset: None if asset is None else asset["name"]
            )

            df_steps["tagValuesLocalisation"] = df_steps["tagValuesLocalisation"].apply(
                lambda full_loc: {
                    loc["tagKey"]["key"]: loc["value"] for loc in full_loc
                }
            )

        if expand_localisation is True and extract_from_localisation is not None:
            df_steps = utils.expand_dataframe_column(
                df_steps, "tagValuesLocalisation", add_prefix=False
            )

        df_steps.set_index(["step_id"], append=True, inplace=True)

        # Format values
        if extract_from_values == "id":
            df_values["values"] = df_values["values"].apply(
                lambda values: {val["data"]["id"]: val["value"] for val in values}
            )
        elif extract_from_values == "reference":
            df_values["values"] = df_values["values"].apply(
                lambda values: {
                    val["data"]["reference"]: val["value"] for val in values
                }
            )
        elif extract_from_values == "description":
            if append_unit_to_description is True:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        f'{val["data"]["description"]} ({val["unit"]["label"]})': val[
                            "value"
                        ]
                        for val in values
                    }
                )
            else:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        val["data"]["description"]: val["value"] for val in values
                    }
                )

        if expand_value_ids is True and extract_from_values is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "values", add_prefix=False
            )

        if extract_from_features == "id":
            df_values["features"] = df_values["features"].apply(
                lambda features: {feat["tagKey"]["id"]: feat["id"] for feat in features}
            )
        elif extract_from_features == "value":
            df_values["features"] = df_values["features"].apply(
                lambda features: {
                    feat["tagKey"]["key"]: feat["value"] for feat in features
                }
            )

        if expand_features is True and extract_from_features is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "features", add_prefix=False
            )

        # Output
        return df_steps, df_values
    else:
        return df


def get_batches(
    batch_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    name: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    asset_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    split_steps_and_values: bool = True,
    extract_from_localisation: Optional[str] = "value",
    expand_localisation: bool = True,
    extract_from_values: Optional[str] = "reference",
    expand_value_ids: bool = True,
    extract_from_features: Optional[str] = "value",
    expand_features: bool = True,
    append_unit_to_description: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get batch instances from the OIAnalytics API

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    name: str, optional
        A string that should be contained by all batch names returned
    features_value_ids: str or list of str, optional
        Possibly multiple feature value ids each returned batch should match.
        If for a given feature multiple feature value ids are provided then a batch will be returned if it
        contains one of them.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    split_steps_and_values: bool, default True
        If True, the response will be split into two separate DataFrames instead of a single dense one.
    extract_from_localisation: {'id', 'value', None}, default 'value'
        What field should be extracted from localisation information. If None, the full dictionary is kept.
    expand_localisation: bool, default True
        Whether or not localisation information should be expanded into multiple columns.
    extract_from_values: {'id', 'reference', 'description', None}, default 'reference'
        What field should be extracted for naming data. If None, the full dictionary is kept.
    expand_value_ids: bool, default True
        Whether or not data should be expanded into multiple columns.
    extract_from_features: {'id', 'value', None}, default 'value'
        What field should be extracted for naming features. If None, the full dictionary is kept.
    expand_features: bool, default True
        Whether or not features should be expanded into multiple columns.
    append_unit_to_description: bool, default True
        Only used when 'extract_from_values' is 'description'. If True, the unit is added after the description.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    If 'split_steps_and_values' is False, a single DataFrame containing batches and their information in dictionaries
    is returned.
    If 'split_steps_and_values' is True (default behaviour), a tuple of 2 DataFrames is returned:
        - Batch steps
        - Batch data and features
    """

    # Args validation
    if extract_from_localisation not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_localisation': {extract_from_localisation}"
        )

    if extract_from_localisation is None and expand_localisation is True:
        raise ValueError(
            "Localisation cannot be expanded if 'extract_from_values' is None"
        )

    if extract_from_values not in ["id", "reference", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_values': {extract_from_values}"
        )

    if extract_from_values is None and expand_value_ids is True:
        raise ValueError("Values cannot be expanded if 'extract_from_values' is None")

    if extract_from_features not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_features': {extract_from_features}"
        )

    if extract_from_features is None and expand_features is True:
        raise ValueError(
            "Features cannot be expanded if 'extract_from_features' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.batches.get_batches(
            batch_type_id=batch_type_id,
            start_date=start_date,
            end_date=end_date,
            name=name,
            features_value_ids=features_value_ids,
            asset_ids=asset_ids,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=["id", "name", "timestamp", "steps", "values", "features"]
            )

        # Rename columns
        page_df.rename(
            columns={
                "id": "batch_id",
                "name": "batch_name",
                "timestamp": "batch_timestamp",
            },
            inplace=True,
        )

        # Parse dates
        page_df["batch_timestamp"] = pd.to_datetime(
            page_df["batch_timestamp"], format="ISO8601"
        )

        # Set index
        page_df.set_index(["batch_id", "batch_name", "batch_timestamp"], inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Split steps and values
    if split_steps_and_values is True:
        # Split dataframe
        df_steps = df.drop(columns=["values", "features"])
        df_values = df.drop(columns="steps")

        # Format steps
        df_steps = df_steps.explode("steps")
        df_steps = utils.expand_dataframe_column(
            df_steps,
            "steps",
            add_prefix=False,
            expected_keys=[
                "step",
                "start",
                "end",
                "localisationType",
                "assetLocalisation",
                "tagValuesLocalisation",
            ],
        )
        df_steps = utils.expand_dataframe_column(
            df_steps, "step", expected_keys=["id", "name"]
        )

        df_steps["start"] = pd.to_datetime(df_steps["start"], format="ISO8601")
        df_steps["end"] = pd.to_datetime(df_steps["end"], format="ISO8601")

        if extract_from_localisation == "id":
            df_steps["assetLocalisation"] = df_steps["assetLocalisation"].map(
                lambda asset: None if asset is None else asset["id"]
            )

            df_steps["tagValuesLocalisation"] = df_steps["tagValuesLocalisation"].apply(
                lambda full_loc: {loc["tagKey"]["id"]: loc["id"] for loc in full_loc}
            )

        elif extract_from_localisation == "value":
            df_steps["assetLocalisation"] = df_steps["assetLocalisation"].map(
                lambda asset: None if asset is None else asset["name"]
            )

            df_steps["tagValuesLocalisation"] = df_steps["tagValuesLocalisation"].apply(
                lambda full_loc: {
                    loc["tagKey"]["key"]: loc["value"] for loc in full_loc
                }
            )

        if expand_localisation is True and extract_from_localisation is not None:
            df_steps = utils.expand_dataframe_column(
                df_steps, "tagValuesLocalisation", add_prefix=False
            )

        df_steps.set_index(["step_id"], append=True, inplace=True)

        # Format values
        if extract_from_values == "id":
            df_values["values"] = df_values["values"].apply(
                lambda values: {val["data"]["id"]: val["value"] for val in values}
            )
        elif extract_from_values == "reference":
            df_values["values"] = df_values["values"].apply(
                lambda values: {
                    val["data"]["reference"]: val["value"] for val in values
                }
            )
        elif extract_from_values == "description":
            if append_unit_to_description is True:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        f'{val["data"]["description"]} ({val["unit"]["label"]})': val[
                            "value"
                        ]
                        for val in values
                    }
                )
            else:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        val["data"]["description"]: val["value"] for val in values
                    }
                )

        if expand_value_ids is True and extract_from_values is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "values", add_prefix=False
            )

        if extract_from_features == "id":
            df_values["features"] = df_values["features"].apply(
                lambda features: {feat["tagKey"]["id"]: feat["id"] for feat in features}
            )
        elif extract_from_features == "value":
            df_values["features"] = df_values["features"].apply(
                lambda features: {
                    feat["tagKey"]["key"]: feat["value"] for feat in features
                }
            )

        if expand_features is True and extract_from_features is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "features", add_prefix=False
            )

        # Output
        return df_steps, df_values
    else:
        return df


def update_batch_values(
    batch_type_id: str,
    data: Union[pd.Series, pd.DataFrame],
    unit_ids: Optional[dict] = None,
    batch_id_index_name: str = "batch_id",
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved
    data: pd.Series or pd.DataFrame
        Object containing the data to be uploaded, where the index contains the batches ids and the
    unit_ids: dict[str, str], optional
        Dictionary with the unit-ids of data to be uploaded.
        Each key is a data-id that points to the corresponding unit-id.
    batch_id_index_name: str, default 'batch_id'
        The name of the Series's or DataFrame's index
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------

    """
    # Init
    if unit_ids is None:
        unit_ids = {}

    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)
    else:
        data = data.copy()

    data.index = data.index.get_level_values(batch_id_index_name)

    # Send each individual value
    def send_value(batch_value_tuple: tuple):
        batch_id = batch_value_tuple[0]
        data_id = batch_value_tuple[1]
        value = batch_value_tuple[2]
        endpoints.batches.update_batch_value(
            batch_type_id=batch_type_id,
            batch_id=batch_id,
            data_id=data_id,
            value=value,
            unit_id=unit_ids.get(data_id, None),
            create_upload_event=create_upload_event,
            api_credentials=api_credentials,
        )

    # Build the iterator over individual batch value tuples (batch_id, data_id, value)
    batch_values = list(
        itertools.chain.from_iterable(
            [
                [(idx,) + i for i in r.iteritems() if not np.isnan(i[1])]
                for idx, r in data.iterrows()
            ]
        )
    )

    with ThreadPoolExecutor() as pool:
        updates = pool.map(send_value, batch_values)

    try:
        print(f"{len(list(updates))} batch values sent to OIAnalytics")
    except Exception:
        print(f"Error during batch values update:\n{traceback.format_exc()}")


def update_batch_feature_values(
    batch_type_id: str,
    data: Union[pd.Series, pd.DataFrame],
    batch_id_index_name: str = "batch_id",
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Insert/update batches features values
    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved.
    data: pd.Series or pd.DataFrame
        Object containing the features values to be uploaded:
        - The index contains the batch-ids;
        - If the object is a DataFrame, the columns contains the features names;
        - If the object is a Series, the Series' name is the feature name being updated
        - The data contains the features values.
    batch_id_index_name: str, default 'batch_id'
        The name of the Dataframe's or Series's index.
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------

    """
    # Init
    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)
    else:
        data = data.copy()

    data.index = data.index.get_level_values(batch_id_index_name)

    # Send each individual value
    def send_value(batch_value_tuple: tuple):
        batch_id = batch_value_tuple[0]
        feature_id = batch_value_tuple[1]
        value = batch_value_tuple[2]
        endpoints.batches.update_batch_feature_value(
            batch_type_id=batch_type_id,
            batch_id=batch_id,
            feature_id=feature_id,
            value=value,
            create_upload_event=create_upload_event,
            api_credentials=api_credentials,
        )

    # Build the iterator over individual batch value tuples (batch_id, data_id, value)
    batch_values = list(
        itertools.chain.from_iterable(
            [
                [(idx,) + i for i in r.iteritems() if not np.isnan(i[1])]
                for idx, r in data.iterrows()
            ]
        )
    )

    with ThreadPoolExecutor() as pool:
        pool.map(send_value, batch_values)


def update_batch_features_and_values(
    batch_type_id: str,
    data: pd.DataFrame,
    feature_columns: Optional[List[str]] = None,
    unit_ids: Optional[dict[str, str]] = None,
    batch_id_index_name: str = "batch_id",
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    It updates many batches values and features values at once

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved.
    data: pd.DataFrame
        Object containing the features values and values to be uploaded:
        - The index contains the batch-ids;
        - The columns contains the features-ids and data-ids;
        - The data contains the features values.
    feature_columns: list of str, optional
        List containing the names of the Dataframe's columns corresponding to the features-ids.
    unit_ids: dict[str, str], optional
        Dictionary with the unit-ids of data to be uploaded.
        Each key is a data-id that points to the corresponding unit-id.
    batch_id_index_name: str, default 'batch_id'
        The name of the Dataframe's or Series's index.
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A dictionary of the response from the API, containing the data insert report
    """

    # Init
    if feature_columns is None:
        feature_columns = []

    if unit_ids is None:
        unit_ids = {}

    data.index = data.index.get_level_values(batch_id_index_name)

    payload = []
    for index in data.index:
        payload.append(
            {
                "batchId": index,
                "batchFeatureCommands": [
                    {
                        "batchTagKeyId": tag,
                        "batchTagValueValue": data.loc[index, tag],
                    }
                    for tag in feature_columns
                    if pd.notnull(data.loc[index, tag])
                ],
                "batchValueCommands": [
                    {
                        "dataId": data_id,
                        "value": float(data.loc[index, data_id]),
                        "unitId": unit_ids.get(data_id),
                    }
                    for data_id in [
                        col for col in data.columns if col not in feature_columns
                    ]
                    if pd.notnull(data.loc[index, data_id])
                ],
            }
        )

    response = endpoints.batches.update_batch_features_and_values(
        batch_type_id=batch_type_id,
        data=payload,
        create_upload_event=create_upload_event,
        api_credentials=api_credentials,
    )

    return response


def update_vector_batch_values(
    data: Union[pd.DataFrame, List[pd.DataFrame]],
    data_reference: Union[str, List[str]],
    index_units: Optional[dict] = None,
    values_units: Optional[dict] = None,
    batch_id_index_name: str = "batch_id",
    use_external_reference: bool = False,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Insert time values stored in a DataFrame through the OIAnalytics API

    Parameters
    ----------
    data: list of dictionaries
        List where each element has a 'dataReference' as a key and a dataframe as a value.
    data_reference: string or list of strings
        The unique data reference for the data being inserted.
    index_units: dictionary, optional
        A dictionary indexed by data reference, specifying the values in which it is sent.
    values_units: dictionary, optional
        A dictionary indexed by data reference, specifying the values in which it is sent.
    batch_id_index_name: string, default 'batch_id'
        The name of the index of the DataFrame(s)
    use_external_reference: bool, default False
        If True, the data are named using their external reference; if False, the OIAnalytics reference is used.
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A dictionary of the response from the API, containing the data insert report
    """

    # Init 'index_units'
    if index_units is None:
        index_units = {}

    # Init 'values_units'
    if values_units is None:
        values_units = {}

    # Init list of data and list of data_reference
    if isinstance(data, pd.DataFrame):
        data = [data.copy()]
        data_reference = [data_reference]

    # Build DTO
    payload = []
    for reference, df in zip(data_reference, data):
        try:
            df.index = df.index.get_level_values(batch_id_index_name)
        except KeyError:
            raise KeyError("The dataframe must have an index level named 'batch_id'")
        # Build payload for the data
        payload.append(
            {
                "dataReference": reference,
                "indexUnit": index_units.get(reference, None),
                "valueUnit": values_units.get(reference, None),
                "batchIds": df.index.tolist(),
                "indexes": df.columns.tolist(),
                "values": df.to_numpy().tolist(),
            }
        )

    # Query endpoint
    response = endpoints.batches.update_vector_batch_values(
        data=payload,
        use_external_reference=use_external_reference,
        create_upload_event=create_upload_event,
        api_credentials=api_credentials,
    )

    # Output
    return response


def update_batch(
    batch_type_id: str,
    batch_id: str,
    name: str,
    steps: pd.DataFrame,
    start_date_name: str = "start",
    end_date_name: str = "end",
    asset_localisation_column: Optional[str] = None,
    tag_localisation_columns: Optional[Union[str, List[str]]] = None,
    features: Optional[pd.DataFrame] = None,
    values: Optional[pd.DataFrame] = None,
    value_column_name: str = "value",
    unit_id_column_name: str = "unit_id",
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing batch.

    Parameters
    ----------
    batch_type_id : str
        The OIAnalytics ID of the batch type the batch belongs to.
    batch_id : str
        The OIAnalytics ID of the batch to update. Must be an existing ID of a batch.
    name : str
        The new name of the batch.
    steps : pd.DataFrame
        The steps of the batches to insert. The DataFrame is indexed by batch name (named after batch_name_index_name).
        The content of this DataFrame must be:
            - The step ID (named after step_id_index_name)
            - The start and end dates of the steps (named after start_date_name and end_date_name)
            - Optional asset or tag localisation columns
    start_date_name : str, default 'start'
        The name of the column containing the start date of the steps id in the steps DataFrame.
    end_date_name : str, default 'end'
        The name of the column containing the end date of the steps id in the steps DataFrame.
    asset_localisation_column : str, optional
        The name of the column containing the asset IDs allowing the localisation of steps.
    tag_localisation_columns : str or list of str, optional
        The name of the columns containing the tag value IDs allowing the localisation of steps.
    features : pd.DataFrame
        Contains the batch feature values to add to the batch.
        It is merged with existing feature values on an already existing feature id will be overwritten.
        The Series must be indexed by batch-tag-key-id and contain the feature value.
    values : pd.DataFrame
        A DataFrame with the values to create on the batch.
        It is merged with existing feature values on an already existing feature id will be overwritten.
        The DataFrame has the following configuration:
        It must be indexes by data ID, and contain a column for the values nad unit-IDs.
        - dataId: str; the ID of the data.
        - value: float | int, optional; the value (if None, the data will be deleted if it previously existed).
        - unitId: str; the ID of the unit the value is expressed in (in its absent, the storage unit of the measurement will be used).
    value_column_name: str, default "value",
        The name of the column in values containing the values.
    unit_id_column_name: str, default "unit_id"
        The name of the column in values containing the units.
    create_upload_event : bool, default True
        Whether to create an upload event with updated values.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API.
        If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    def build_single_step_payload(step_id, step):
        # Step dates
        step_payload = {
            "stepId": step_id,
            "start": utils.get_zulu_isoformat(step[start_date_name]),
            "end": utils.get_zulu_isoformat(step[end_date_name]),
        }

        # Localisation
        if (
            asset_localisation_column is not None
            and step.notna()[asset_localisation_column]
        ):
            step_payload["localisationType"] = "ASSET"
            step_payload["localisationTagValueIds"] = []
            step_payload["localisationAssetId"] = step[asset_localisation_column]

        elif (
            tag_localisation_columns is not None
            and len(tag_localisation_columns) > 0
            and step.notna()[tag_localisation_columns].any()
        ):
            step_payload["localisationType"] = "TAG_VALUES"
            step_payload["localisationTagValueIds"] = (
                step[tag_localisation_columns].dropna().to_list()
            )
            step_payload["localisationAssetId"] = None

        else:
            step_payload["localisationType"] = "NO_LOCALISATION"
            step_payload["localisationTagValueIds"] = []
            step_payload["localisationAssetId"] = None

        return step_payload

    steps_payload = [build_single_step_payload(step_id, step) for step_id, step in steps.iterrows()]

    values_payload = [
        {
            "dataId": data_id,
            "value": value[value_column_name],
            "unitId": value[unit_id_column_name],
        }
        for data_id, value in values.iterrows()
    ]

    batch_features = [
        {"batchTagKeyId": index, "batchTagValueValue": value}
        for index, value in features.items()
    ]

    # endpoint
    response = endpoints.batches.update_batch(
        batch_type_id=batch_type_id,
        batch_id=batch_id,
        name=name,
        steps=steps_payload,
        tag_values_by_id=[],
        tag_values_by_values=batch_features,
        values=values_payload,
        create_upload_event=create_upload_event,
        api_credentials=api_credentials,
    )

    # Output
    return response


def create_or_update_batches(
    batch_type_id: str,
    steps: pd.DataFrame,
    values: Optional[pd.DataFrame] = None,
    values_unit_ids: Optional[dict] = None,
    features: Optional[pd.DataFrame] = None,
    vector_data_values: Optional[Union[pd.DataFrame, List[pd.DataFrame]]] = None,
    vector_data_references: Optional[List[str]] = None,
    vector_data_index_units: Optional[dict] = None,
    vector_data_values_units: Optional[dict] = None,
    on_duplicates_keep: str = "last",
    batch_name_index_name: str = "batch_name",
    step_id_index_name: str = "step_id",
    start_date_name: str = "start",
    end_date_name: str = "end",
    asset_localisation_column: Optional[str] = None,
    tag_localisation_columns: Optional[Union[str, List[str]]] = None,
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Create or update batches, including steps, values, vector values and features

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved.
    steps: pd.DataFrame
        The steps of the batches to insert. The DataFrame is indexed by batch name (named after batch_name_index_name).
        The content of this DataFrame must be:
            - The step ID (named after step_id_index_name)
            - The start and end dates of the steps (named after start_date_name and end_date_name)
            - Optional asset or tag localisation columns
    values: pd.DataFrame, optional
        The data values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
        Each column is named after the data ID.
    values_unit_ids: dict, optional
        A dictionary where keys are data IDs and values are the corresponding unit IDs
    features: pd.DataFrame, optional
        The feature values to update or insert. The DataFrame must be indexed by batch name (Cf. batch_name_index_name).
        Each column is named after the feature ID.
    vector_data_values: pd.DataFrame or list of pd.DataFrame, optional
        The vector data values to insert or update. Must be indexed by batch name (Cf. batch_name_index_name).
        Each column is a vector index.
    vector_data_references: list of str, optional
        The list of the vector data references matching the list  of vector data values DataFrames
    vector_data_index_units: dict, optional
        A dictionary where keys are vector data references and values are the corresponding unit labels for indexes
    vector_data_values_units: dict, optional
        A dictionary where keys are vector data references and values are the corresponding unit labels for values
    on_duplicates_keep: {'first', 'last', False}, default 'last'
        Indicate which DataFrame row to keep in case of duplicates (Cf. pd.DataFrame.drop_duplicates)
    batch_name_index_name: str, default 'batch_name'
        The name of the index level containing the batch name in the DataFrames
    step_id_index_name: str, default 'step_id'
        The name of the index level (or column) containing the step id in the steps DataFrame
    start_date_name: str, default 'start'
        The name of the column containing the start date of the steps id in the steps DataFrame
    end_date_name: str, default 'end'
        The name of the column containing the end date of the steps id in the steps DataFrame
    asset_localisation_column: str, optional
        The name of the column containing the asset IDs allowing the localisation of steps
    tag_localisation_columns: str or list of str, optional
        The name of the columns containing the tag value IDs allowing the localisation of steps
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A dictionary of the response from the API, containing the list of created or updated batches
    """

    # Validate arguments
    if on_duplicates_keep not in ["first", "last"]:
        raise ValueError(
            f"Unexpected value for 'on_duplicates_keep': {on_duplicates_keep}. Only 'first' and 'last' are accepted."
        )

    # Init
    if values_unit_ids is None:
        values_unit_ids = {}

    if isinstance(tag_localisation_columns, str):
        tag_localisation_columns = [tag_localisation_columns]

    if isinstance(vector_data_values, pd.DataFrame):
        vector_data_values = [vector_data_values]

    # Build payload
    batches = []

    # Append each batch to the payload
    for batch_name in steps.index.get_level_values(batch_name_index_name).unique():
        # Init
        batch_command = {"name": batch_name}

        # Build steps command
        batch_steps = steps.loc[
            steps.index.get_level_values(batch_name_index_name) == batch_name
        ].reset_index()

        def build_single_step_payload(step):
            # Step dates
            step_payload = {
                "stepId": step[step_id_index_name],
                "start": utils.get_zulu_isoformat(step[start_date_name]),
                "end": utils.get_zulu_isoformat(step[end_date_name]),
            }

            # Localisation
            if (
                asset_localisation_column is not None
                and step.notna()[asset_localisation_column]
            ):
                step_payload["localisationType"] = "ASSET"
                step_payload["localisationTagValueIds"] = []
                step_payload["localisationAssetId"] = step[asset_localisation_column]

            elif (
                tag_localisation_columns is not None
                and len(tag_localisation_columns) > 0
                and step.notna()[tag_localisation_columns].any()
            ):
                step_payload["localisationType"] = "TAG_VALUES"
                step_payload["localisationTagValueIds"] = (
                    step[tag_localisation_columns].dropna().to_list()
                )
                step_payload["localisationAssetId"] = None

            else:
                step_payload["localisationType"] = "NO_LOCALISATION"
                step_payload["localisationTagValueIds"] = []
                step_payload["localisationAssetId"] = None

            return step_payload

        batch_command["steps"] = batch_steps.apply(
            build_single_step_payload, axis=1
        ).to_list()

        # Build values command
        if values is None:
            batch_command["values"] = []
        else:
            batch_values = values.loc[
                (values.index.get_level_values(batch_name_index_name) == batch_name)
                & (~values.index.duplicated(keep=on_duplicates_keep))
            ]

            if batch_values.shape[0] == 0:
                batch_command["values"] = []
            else:

                def build_single_value_payload(value):
                    return {
                        "dataId": value.name,
                        "value": float(value.iloc[0]),
                        "unitId": values_unit_ids.get(value.name),
                    }

                batch_command["values"] = (
                    batch_values.dropna().apply(build_single_value_payload).to_list()
                )

        # Build features command
        batch_command["tagValuesById"] = []  # Not used

        if features is None:
            batch_command["tagValuesByValue"] = []
        else:
            batch_features = features.loc[
                features.index.get_level_values(batch_name_index_name) == batch_name
            ].drop_duplicates(keep=on_duplicates_keep)

            if batch_features.shape[0] == 0:
                batch_command["tagValuesByValue"] = []
            else:

                def build_single_feature_payload(feature):
                    return {
                        "batchTagKeyId": feature.name,
                        "batchTagValueValue": feature.iloc[0],
                    }

                batch_command["tagValuesByValue"] = (
                    batch_features.dropna()
                    .apply(build_single_feature_payload)
                    .to_list()
                )

        # Append batch command
        batches.append(batch_command)

    # Query endpoint
    response = endpoints.batches.create_or_update_batches(
        batch_type_id=batch_type_id,
        batches=batches,
        create_upload_event=create_upload_event,
        api_credentials=api_credentials,
    )

    # Update vector data values
    if vector_data_values is not None:
        batch_ids_renaming_dict = {b["name"]: b["id"] for b in response}

        # Rename batches to IDs
        vector_data_values_by_batch_id = []

        for vector_data_df in vector_data_values:
            vector_data_df_index = vector_data_df.index.to_frame()
            vector_data_df_index[batch_name_index_name] = vector_data_df_index[
                batch_name_index_name
            ].map(batch_ids_renaming_dict)
            vector_data_df_by_batch_id = vector_data_df.copy()
            vector_data_df_by_batch_id.index = pd.MultiIndex.from_frame(
                vector_data_df_index
            )
            vector_data_values_by_batch_id.append(
                vector_data_df_by_batch_id.loc[
                    ~vector_data_df_by_batch_id.index.duplicated(
                        keep=on_duplicates_keep
                    )
                ]
            )

        # Actually send vector data values
        update_vector_batch_values(
            data=vector_data_values_by_batch_id,
            data_reference=vector_data_references,
            index_units=vector_data_index_units,
            values_units=vector_data_values_units,
            batch_id_index_name=batch_name_index_name,
            create_upload_event=create_upload_event,
        )

    # Output
    return response


# Batch Relations
def get_batch_relations(
    batch_structure_relation_id: Optional[str] = None,
    start_date: Optional[Union[str, datetime]] = None,
    end_date: Optional[Union[str, datetime]] = None,
    source_batch_name: Optional[str] = None,
    destination_batch_name: Optional[str] = None,
    expand_batch_structure_relation: bool = True,
    expand_source_batch: bool = True,
    expand_destination_batch: bool = True,
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    Search batch relations in OIAnalytics by criteria.

    Parameters
    ----------
    batch_structure_relation_id : str
        The OIAnalytics ID of the batch structure relation.
    start_date : Union[str, datetime], optional
        The start of the interval used to find relations.
        If passed as a string, it should be in ISO 8601 format.
    end_date : Union[str, datetime], optional
        The end of the interval used to find relations.
        If passed as a string, it should be in ISO 8601 format.
    source_batch_name : str, optional
        The name of the source batch.
    destination_batch_name : str, optional
        The name of the destination batch.
    expand_batch_structure_relation : bool, default True
        Whether to expand the 'batchStructureRelation' column.
        If True, it is split into two columns: 'batchStructureRelation_id' and 'batchStructureRelation_name'.
    expand_source_batch : bool, default True
        Whether to expand the 'sourceBatch' column.
        If True, it is split into two columns: 'sourceBatch_id' and 'sourceBatch_name'.
    expand_destination_batch : bool, default True
        Whether to expand the 'destinationBatch' column.
        If True, it is split into two columns: 'destinationBatch_id' and 'destinationBatch_name'.
    page : int, default 0
        The page to retrieve, defaults to 0 e.g. the first page.
    page_size : int, default 20
        The size of a page, defaults to 20.
    get_all_pages : bool, True
        Whether to get batch relations of all pages.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials : _credentials.OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the batch relations.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    if get_all_pages:
        page = 0
        page_size = 500

    def get_page(page_num: int) -> dict:
        return endpoints.batches.get_batch_relations(
            batch_structure_relation_id=batch_structure_relation_id,
            start_date=start_date,
            end_date=end_date,
            source_batch_name=source_batch_name,
            destination_batch_name=destination_batch_name,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict) -> pd.DataFrame:
        page_df = pd.DataFrame.from_dict(page_response["content"])
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "batchStructureRelation",
                    "sourceBatch",
                    "destinationBatch",
                    "weight",
                ]
            )
        page_df.set_index("id", inplace=True)
        return page_df

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    if expand_batch_structure_relation:
        df = utils.expand_dataframe_column(
            df, "batchStructureRelation", add_prefix=True
        )

    if expand_source_batch:
        df = utils.expand_dataframe_column(df, "sourceBatch", add_prefix=True)

    if expand_destination_batch:
        df = utils.expand_dataframe_column(df, "destinationBatch", add_prefix=True)

    # Output
    return df


def get_single_batch_relation(
    batch_relation_id: Optional[str] = None,
    expand_batch_structure_relation: bool = True,
    expand_source_batch: bool = True,
    expand_destination_batch: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Get the detail of a batch relation by its OIAnalytics ID.

    Parameters
    ----------
    batch_relation_id : str
        The OIAnalytics ID of the batch relation to retrieve.
    expand_batch_structure_relation : bool, default True
        Whether to expand the 'batchStructureRelation' column.
        If True, it is split into two columns: 'batchStructureRelation_id' and 'batchStructureRelation_name'.
    expand_source_batch : bool, default True
        Whether to expand the 'sourceBatch' column.
        If True, it is split into two columns: 'sourceBatch_id' and 'sourceBatch_name'.
    expand_destination_batch : bool, default True.
        Whether to expand the 'destinationBatch' column.
        If True, it is split into two columns: 'destinationBatch_id' and 'destinationBatch_name'.
    api_credentials : _credentials.OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series containing the detail of a batch relation.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.batches.get_single_batch_relation(
        batch_relation_id=batch_relation_id, api_credentials=api_credentials
    )

    df = pd.DataFrame([response])

    # check if there are relations
    if len(df) == 0:
        # Output
        return df.squeeze()

    df.set_index("id", inplace=True)

    if expand_batch_structure_relation:
        df = utils.expand_dataframe_column(
            df, "batchStructureRelation", add_prefix=True
        )

    if expand_source_batch:
        df = utils.expand_dataframe_column(df, "sourceBatch", add_prefix=True)

    if expand_destination_batch:
        df = utils.expand_dataframe_column(df, "destinationBatch", add_prefix=True)

    ser = df.squeeze()

    # Output
    return ser



from typing import Optional, Union, List, Literal
from datetime import datetime

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = [
    "get_event_types",
    "get_event_type_details",
    "create_event_type",
    "get_events",
    "create_or_update_events",
]


def get_event_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    extract_from_tag_keys: Optional[str] = "value",
    extract_from_context: Optional[str] = "value",
    expand_context: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the configured event types from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    extract_from_tag_keys: {'id', 'value', 'dict', None}, default 'value'
        What field should be extracted from tag keys information.
        If 'dict', a dictionary in the form of {id: value} is built. If None, the full dictionary is kept.
    extract_from_context: {'id', 'value', None}, default 'value'
        What field should be extracted from context information. If None, the full dictionary is kept.
    expand_context: bool, default True
        Whether or not the context should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pandas.DataFrame
        A DataFrame listing event types.
    """

    # Args validation
    if extract_from_tag_keys not in ["id", "value", "dict", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_tag_keys': {extract_from_tag_keys}"
        )

    if extract_from_context not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_context': {extract_from_context}"
        )

    if expand_context is True and extract_from_context is None:
        raise ValueError("Context cannot be expanded if 'extract_from_context' is None")

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.events.get_event_types(
            page=page_num, page_size=page_size, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name", "tagKeys", "tagContext"])

        # Extract from tag keys
        if extract_from_tag_keys == "id":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: [tk["id"] for tk in l]
            )
        elif extract_from_tag_keys == "value":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: [tk["key"] for tk in l]
            )
        elif extract_from_tag_keys == "dict":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: {tk["id"]: tk["key"] for tk in l}
            )

        # Extract from context
        if extract_from_context == "id":
            page_df["tagContext"] = page_df["tagContext"].apply(
                lambda context: {
                    c["tagKey"]["id"]: [atv["id"] for atv in c["accessibleTagValues"]]
                    for c in context
                }
            )
        elif extract_from_context == "value":
            page_df["tagContext"] = page_df["tagContext"].apply(
                lambda context: {
                    c["tagKey"]["key"]: [
                        atv["value"] for atv in c["accessibleTagValues"]
                    ]
                    for c in context
                }
            )

        if expand_context is True and extract_from_context is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "tagContext", add_prefix=False
            )

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_event_type_details(
    event_type_id: str,
    extract_from_tag_keys: Literal["id", "key", None] = "key",
    extract_from_asset_types: Literal["id", "name", None] = "name",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Get the detail of an event type by its OIAnalytics ID.

    Parameters
    ----------
    event_type_id : str
        The ID of the event type to retrieve.
    extract_from_tag_keys : {'id', 'key', None}, default 'key'
        Key to be associated from the column 'tagKeys'.
    extract_from_asset_types : {'id', 'name', None}, default 'name'
        Key to be associated from the column 'assetTypes'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the event type.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    dict_response = endpoints.events.get_event_type_details(
        event_type_id=event_type_id,
        api_credentials=api_credentials,
    )

    ser = pd.Series(dict_response, dtype=object)

    if extract_from_tag_keys is not None:
        ser.loc["tagKeys"] = [
            tag_key[extract_from_tag_keys] for tag_key in ser.loc["tagKeys"]
        ]

    if extract_from_asset_types is not None:
        ser.loc["assetTypes"] = [
            tag_key[extract_from_asset_types] for tag_key in ser.loc["assetTypes"]
        ]

    # Output
    return ser


def create_event_type(
    name: str,
    asset_type_ids: List[str],
    tag_keys: List[dict],
    tag_context: List[dict],
    extract_from_tag_keys: Literal["id", "key", None] = "key",
    extract_from_asset_types: Literal["id", "name", None] = "name",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create an event type.

    Parameters
    ----------
    name : str
        The name of the event type to create.
    asset_type_ids : list of str
        The IDs of the types of asset that can fire this type of event.
    tag_keys : list of dict
        The array of source tag keys to create on the event type.
    tag_context : list of dict
        The array of tag access to attach to the event type to restrict access of users.
    extract_from_tag_keys : {'id', 'key', None}, default 'key'
        Key to be associated from the column 'tagKeys'.
    extract_from_asset_types : {'id', 'name', None}, default 'name'
        Key to be associated from the column 'assetTypes'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the event that has been created.
    """

    dict_response = endpoints.events.create_event_type(
        name=name,
        asset_type_ids=asset_type_ids,
        tag_keys=tag_keys,
        tag_context=tag_context,
        api_credentials=api_credentials,
    )

    ser = pd.Series(dict_response, dtype=object)

    if extract_from_tag_keys is not None:
        ser.loc["tagKeys"] = [
            tag_key[extract_from_tag_keys] for tag_key in ser.loc["tagKeys"]
        ]

    if extract_from_asset_types is not None:
        ser.loc["assetTypes"] = [
            tag_key[extract_from_asset_types] for tag_key in ser.loc["assetTypes"]
        ]

    # Output
    return ser


def get_events(
    event_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    description: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_event_type: bool = True,
    extract_from_features: Optional[str] = "key",
    expand_features: bool = True,
    extract_from_values: Optional[str] = "reference",
    append_unit_to_description: bool = True,
    extract_from_assets: Optional[str] = "name",
    expand_values: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get event instances from the OIAnalytics API

    Parameters
    ----------
    event_type_id: str
        The id of the event type to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    description: str, optional
        A string that should be contained by all events description returned
    features_value_ids: str or list of str, optional
        Possibly multiple feature value ids each returned event should match.
        If for a given feature multiple feature value ids are provided than a event will be returned if it
        contains one of them.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_event_type: bool, default True
        Whether or not the event type information should be expanded into multiple columns.
    extract_from_features: {'id', 'value', None}, default 'value'
        What field should be extracted from features information. If None, the full dictionary is kept.
    expand_features: bool, default True
        Whether or not the features should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing event dates, data and features values.
    """

    # Args validation
    if extract_from_features not in ["id", "key", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_features': {extract_from_features}"
        )

    if expand_features is True and extract_from_features is None:
        raise ValueError(
            "Features cannot be expanded if 'extract_from_features' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.events.get_events(
            event_type_id=event_type_id,
            start_date=start_date,
            end_date=end_date,
            description=description,
            features_value_ids=features_value_ids,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "eventType",
                    "description",
                    "start",
                    "end",
                    "duration",
                    "tagValues",
                    "values",
                    "assets",
                ]
            )

        # Format dataframe
        if expand_event_type is True:
            page_df = utils.expand_dataframe_column(
                page_df, "eventType", expected_keys=["id", "name"]
            )

        if extract_from_features == "id":
            page_df["tagValues"] = page_df["tagValues"].apply(
                lambda fl: {f["eventTagKey"]["id"]: f["value"] for f in fl}
            )
        elif extract_from_features == "key":
            page_df["tagValues"] = page_df["tagValues"].apply(
                lambda fl: {f["eventTagKey"]["key"]: f["value"] for f in fl}
            )

        if expand_features is True and extract_from_features is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "tagValues", add_prefix=False
            )

        # Format values
        if extract_from_values == "id":
            page_df["values"] = page_df["values"].apply(
                lambda values: {val["data"]["id"]: val["value"] for val in values}
            )
        elif extract_from_values == "reference":
            page_df["values"] = page_df["values"].apply(
                lambda values: {
                    val["data"]["reference"]: val["value"] for val in values
                }
            )
        elif extract_from_values == "description":
            if append_unit_to_description is True:
                page_df["values"] = page_df["values"].apply(
                    lambda values: {
                        f'{val["data"]["description"]} ({val["unit"]["label"]})': val[
                            "value"
                        ]
                        for val in values
                    }
                )
            else:
                page_df["values"] = page_df["values"].apply(
                    lambda values: {
                        val["data"]["description"]: val["value"] for val in values
                    }
                )

        if expand_values is True and extract_from_values is not None:
            page_df = utils.expand_dataframe_column(page_df, "values", add_prefix=False)

        # Format assets
        if extract_from_assets is not None:
            page_df["assets"] = page_df["assets"].apply(
                lambda assets: [asset[extract_from_assets] for asset in assets]
            )

        page_df.set_index("id", inplace=True)

        # Output
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Dates
    df["start"] = pd.to_datetime(df["start"], format="ISO8601")
    df["end"] = pd.to_datetime(df["end"], format="ISO8601")

    # Output
    return df


def create_or_update_events(
    event_type_id: str,
    events: pd.DataFrame,
    value_columns: List[str] = [],
    feature_columns: List[str] = [],
    asset_column: Optional[List[dict]] = None,
    values_unit_ids: Optional[dict] = None,
    start_date_name: str = "start",
    end_date_name: str = "end",
    description_name: str = "description",
    create_upload_event: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Create or update batches, including steps, values, vector values and features

    Parameters
    ----------
    event_type_id: str
        The id of the batch type to be retrieved.
    events: pd.DataFrame
        The values to update or insert. The DataFrame must be indexed by index id.
        Each column is named after the data or feature ID.
    value_columns: list of strings, default []
        The list of columns containing event data values.
    feature_columns: list of strings, default []
        The list of columns containing event feature values.
    asset_column: str, optional, default None
        The name of the column containing assets.
    values_unit_ids: dict, optional
        A dictionary where keys are data IDs and values are the corresponding unit IDs
    start_date_name: str, default 'start'
        The name of the column containing the start date of the steps id in the steps DataFrame
    end_date_name: str, default 'end'
        The name of the column containing the end date of the steps id in the steps DataFrame
    description_name: string, default 'description'
        The name of the column containing the events description
    create_upload_event: bool, default True
        Whether to create an upload event with updated values.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A dictionary of the response from the API, containing the list of created or updated events
    """

    # Init
    if values_unit_ids is None:
        values_unit_ids = {}

    # Serialization function
    def build_single_event_payload(event: pd.Series):
        # Init
        event_command = {
            "id": None if pd.isna(event.name) else event.name,
            "start": utils.get_zulu_isoformat(event[start_date_name]),
            "end": utils.get_zulu_isoformat(event[end_date_name]),
            "description": event[description_name],
            "tagValues": [
                {"tagKeyId": feat, "tagValueValue": event[feat]}
                for feat in feature_columns
            ],
            "values": [
                {
                    "dataId": val,
                    "value": None if pd.isna(event[val]) else event[val],
                    "unitId": values_unit_ids.get(val),
                }
                for val in value_columns
            ],
            "assetIds": None if asset_column is None else event[asset_column],
        }

        return event_command

    # Build payload
    payload = events.apply(build_single_event_payload, axis=1).to_list()

    # Query endpoint
    response = endpoints.events.create_or_update_events(
        event_type_id=event_type_id,
        events=payload,
        create_upload_event=create_upload_event,
        api_credentials=api_credentials,
    )

    # Output
    return response



from typing import Optional, List
import pandas as pd

from .. import _credentials
from .. import endpoints


__all__ = [
    "get_permissions",
    "get_profiles",
    "get_single_profile",
    "create_profile",
]


def get_permissions(
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List all permissions details.

    Parameters
    ----------
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.DataFrame
        DataFrame containing all the permissions and their details.
        Index:
        - 'name': the name of the permission.
        Columns:
        - 'group': The group this permission belongs to.
        - 'type': The type of permission, can be yes/no or read/write.
        - 'subGroup': The sub-group this permission belongs to.
        - 'operation': The operation attached to the permission.
    """
    response = endpoints.profiles.get_permissions(
        api_credentials=api_credentials,
    )
    df = pd.DataFrame(response)
    if len(df) > 0:
        df.set_index("name", inplace=True)

    # Output
    return df


def get_profiles(
    page: int = 0,
    extract_mode: bool = False,
    permission_name: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List profiles by page.

    Parameters
    ----------
    page : int, default 0.
        The page to retrieve, defaults to 0 e.g. the first page.
    extract_mode : bool, default False
        Whether to extract the key 'mode' for the requested permission name.
        Used only if permission_to_extract is not None.
    permission_name : str, optional
        The name of the permission whose mode is to be retrieved.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame with the detail of the profiles at a given page.
        Index:
        - 'id': the OIAnalytics ID of the profile.
        Columns:
        - 'name': The name of the profile.
        - 'externalId': The external ID of the profile.
        - 'accessRights': The detail of all the permissions of the profile.
    """
    if extract_mode is True and permission_name is None:
        raise ValueError("permission_name cannot be None if extract_mode is True")

    response = endpoints.profiles.get_profiles(
        page=page,
        api_credentials=api_credentials,
    )
    df = pd.DataFrame(response["content"])

    if len(df) == 0:
        return df

    df.set_index("id", inplace=True)

    if extract_mode:
        df["accessRights"] = df["accessRights"].apply(
            lambda x: pd.DataFrame(x)
            .set_index("permission")
            .loc[permission_name, "mode"]
        )
        df = df.rename(columns={"accessRights": f"{permission_name}_accessRights_mode"})

    # Output
    return df


def get_single_profile(
    profile_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the access rights associated with to a profile ID.

    Parameters
    ----------
    profile_id : str
        The OIAnalytics ID of the profile to retrieve.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas Series mapping the permissions and their modes.
    """
    response = endpoints.profiles.get_single_profile(
        profile_id=profile_id, api_credentials=api_credentials
    )
    df = pd.DataFrame(response["accessRights"])

    df.set_index("permission", inplace=True)

    # Output
    return df.squeeze()


def create_profile(
    name: str,
    external_id: Optional[str],
    access_rights: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new profile.

    Parameters
    ----------
    name : str
        The name of the profile.
    external_id : str
        The ID of a group in identity provider to associate with the profile.
    access_rights : list of dict
        The access rights associated with this profile.
        Each dictionary in the list contains the following keys:
        - 'permission': str, the permission the access right defines.
        - 'mode': {'YES', 'NO','READ',WRITE'}, the mode of access right is permitted.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the profile that has just been created.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # endpoints
    response = endpoints.profiles.create_profile(
        name=name,
        external_id=external_id,
        access_rights=access_rights,
        api_credentials=api_credentials,
    )

    ser = pd.Series(response, dtype=object)

    return ser



from typing import Optional, List, Union
from datetime import datetime

import requests

from .. import _credentials
from .. import utils


__all__ = [
    "create_temporal_computation_jobs",
    "create_batch_computation_jobs",
    "create_event_computation_jobs",
]


def create_temporal_computation_jobs(
    data_id: str,
    start_date: Union[str, datetime],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/computation-jobs/continuous-data"
    response = requests.post(
        url=url,
        json={"dataId": data_id, "start": start_date_iso},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def create_batch_computation_jobs(
    batch_type_id: str,
    data_ids: Optional[List[str]] = None,
    batch_ids: Optional[List[str]] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/computation-jobs/batch-data"
    response = requests.post(
        url=url,
        json={
            "batchStructureId": batch_type_id,
            "dataIds": data_ids,
            "batchIds": batch_ids,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def create_event_computation_jobs(
    event_type_id: str,
    data_ids: Optional[List[str]] = None,
    event_ids: Optional[List[str]] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/computation-jobs/event-data"
    response = requests.post(
        url=url,
        json={
            "eventTypeId": event_type_id,
            "dataIds": data_ids,
            "eventIds": event_ids,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def toto():
    pass



import io
from typing import Callable, Optional, Any

from .. import _dtos, outputs
from ...api import endpoints

__all__ = [
    "test_load_resources",
    "test_process_file",
    "test_full_model",
    "get_file_content",
]


def test_load_resources(
    load_resources: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_resources function
    return load_resources(**model_execution.execution_kwargs)


def test_process_file(
    process_file: Callable,
    file_content: io.BytesIO,
    resources: Optional[Any] = None,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run process_file function
    model_output = process_file(
        file_content=file_content,
        resources=resources,
        **model_execution.execution_kwargs
    )

    if not isinstance(
        model_output,
        (
            outputs.OIModelOutputs,
            outputs.FileOutput,
            outputs.TimeValuesOutput,
            outputs.VectorTimeValuesOutput,
            outputs.BatchValuesOutput,
            outputs.VectorBatchValuesOutput,
            outputs.BatchFeaturesOutput,
            outputs.CustomJsonOutput,
            outputs.CustomTextOutput,
            outputs.Delay,
        ),
    ):
        print(
            "Warning: the model output should be an instance of an output class from oianalytics.models.outputs"
        )

    return model_output


def test_full_model(
    load_resources: Callable,
    process_file: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Load file content
    file_content = endpoints.files.get_file_from_file_upload(
        file_upload_id=model_execution.executionContext.uploadEventId,
    )

    # Run load_resources function
    resources = load_resources(**model_execution.execution_kwargs)

    model_output = test_process_file(
        process_file=process_file,
        file_content=file_content,
        resources=resources,
    )

    return model_output


def get_file_content(model_execution: Optional[_dtos.ModelExecution] = None):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()
    # call the endpoint
    return endpoints.files.get_file_from_file_upload(
        file_upload_id=model_execution.executionContext.uploadEventId,
    )



from typing import Optional
import requests

from .. import _credentials


__all_ = ["get_units", "get_unit_families"]


def get_units(
    label: Optional[str] = None,
    quantity_id: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/units"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
            "label": label,
            "quantityId": quantity_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_unit_families(
    label: Optional[str] = None,
    main_unit_id: Optional[str] = None,
    quantity_id: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/unit-families"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
            "label": label,
            "mainUnitId": main_unit_id,
            "quantityId": quantity_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional, Any, Callable

import pandas as pd

from .. import outputs
from .. import _dtos
from .. import get_single_obs_batch_data, get_single_obs_time_data

__all__ = [
    "get_input_data",
    "test_load_resources",
    "test_process_data",
    "test_full_model",
]


def get_input_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Get data
    if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
        raise ValueError(
            "The provided model execution input is not in single observation mode"
        )

    if model_execution.pythonModelInstance.singleObservationContext.type == "time":
        return get_single_obs_time_data(
            model_execution=model_execution,
            rename_data_to_source_code_name=rename_data_to_source_code_name,
        )

    elif model_execution.pythonModelInstance.singleObservationContext.type == "batch":
        return get_single_obs_batch_data(
            model_execution=model_execution,
            rename_data_to_source_code_name=rename_data_to_source_code_name,
        )


def test_load_resources(
    load_resources: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_resources function
    return load_resources(**model_execution.execution_kwargs)


def test_process_data(
    process_data: Callable,
    data: pd.DataFrame,
    resources: Optional[Any] = None,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Init outputs storage
    model_outputs = outputs.OIModelOutputs()

    # Run process_data function
    output_data = data.apply(
        lambda r: pd.Series(process_data(resources=resources, **r.to_dict())),
        axis=1,
    )

    # list of time values sources codes names
    time_values_names = [
        col
        for col in output_data.columns
        if col
        in model_execution.get_data_output_dict(
            data_type="time", values_type="scalar", mode="reference"
        ).keys()
    ]

    # add time values output object
    if len(time_values_names) > 0:
        model_outputs.add_output(
            outputs.TimeValuesOutput(
                data=output_data[time_values_names],
            )
        )

    # list of vector time values sources codes names
    vector_time_values_names = [
        col
        for col in output_data.columns
        if col
        in model_execution.get_data_output_dict(
            data_type="time", values_type="vector", mode="object"
        ).keys()
    ]
    output_vector_time_values = output_data[vector_time_values_names]

    if len(vector_time_values_names) > 0:
        df_list = [
            pd.concat(
                output_vector_time_values[name].to_dict(),
                axis=1,
                sort=True,
            ).T
            for name in vector_time_values_names
        ]

        # create vector time values output object
        model_outputs.add_output(
            outputs.VectorTimeValuesOutput(
                data=df_list,
                data_reference=vector_time_values_names,
            )
        )

    single_obs_context_type = (
        model_execution.pythonModelInstance.singleObservationContext.type
    )

    if len(vector_time_values_names) > 0:
        df_list = [
            pd.concat(
                output_vector_time_values[name].to_dict(),
                axis=1,
                sort=True,
            ).T
            for name in vector_time_values_names
        ]

        # create vector time values output object
        model_outputs.add_output(
            outputs.VectorTimeValuesOutput(
                data=df_list,
                data_reference=vector_time_values_names,
            )
        )

    if single_obs_context_type == "batch":
        # list of batch values sources codes names
        batch_values_names = [
            col
            for col in output_data.columns
            if col
            in model_execution.get_data_output_dict(
                data_type="batch", values_type="scalar", mode="id"
            ).keys()
        ]

        if len(batch_values_names) > 0:
            model_outputs.add_output(
                outputs.BatchValuesOutput(
                    batch_type_id=model_execution.pythonModelInstance.singleObservationContext.batchPredicate.batchType.id,
                    data=output_data[batch_values_names],
                )
            )

        # list of vector batch values sources codes names
        vector_batch_values_names = [
            col
            for col in output_data.columns
            if col
            in model_execution.get_data_output_dict(
                data_type="batch", values_type="vector", mode="object"
            ).keys()
        ]
        # DataFrame containing only with vector batch values
        output_vector_batch_values = output_data[vector_batch_values_names]

        if len(vector_batch_values_names) > 0:
            df_list = [
                pd.concat(
                    output_vector_batch_values[name].to_dict(),
                    axis=1,
                    sort=True,
                ).T
                for name in vector_batch_values_names
            ]

            # create output object for batch vector values
            model_outputs.add_output(
                outputs.VectorBatchValuesOutput(
                    data_reference=vector_batch_values_names,
                    data=df_list,
                )
            )

        # list with batch features names
        features_names = [
            col
            for col in output_data.columns
            if col
            in model_execution.get_output_dict(output_types=["BATCH_TAG_KEY"]).keys()
        ]
        # DataFrame only with batches features
        output_features = output_data[features_names]

        # create output features object
        if len(features_names) > 0:
            model_outputs.add_output(
                outputs.BatchFeaturesOutput(
                    batch_type_id=model_execution.pythonModelInstance.singleObservationContext.batchPredicate.batchType.id,
                    data=output_features,
                )
            )

    # Output
    return model_outputs


def test_full_model(
    load_resources: Callable,
    process_data: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Load data
    data = get_input_data(model_execution=model_execution)

    # Load resources
    resources = test_load_resources(
        load_resources=load_resources, model_execution=model_execution
    )

    # Process data
    return test_process_data(
        process_data=process_data,
        data=data,
        resources=resources,
        model_execution=model_execution,
    )



from typing import Optional, List, Dict, Any

import requests

from .. import _credentials


__all__ = [
    "get_asset_types",
    "get_single_asset_type",
    "delete_asset_type",
    "create_asset_type",
    "update_asset_type",
    "get_assets",
    "get_single_asset",
    "update_single_asset_tags_and_values",
    "delete_asset",
    "create_asset",
    "update_asset",
]


def get_asset_types(
    page: Optional[int] = None,
    name: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/asset-types"
    response = requests.get(
        url=url,
        params={"page": page, "name": name},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_asset_type(
    asset_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/asset-types/{asset_type_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def delete_asset_type(
    asset_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an asset type identified by its ID. The deletion will fail if the asset type is used somewhere in the system.

    Parameters
    ----------
    asset_type_id : str
        The OIAnalytics ID of the asset type to retrieve.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/asset-types/{asset_type_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def create_asset_type(
    name: str,
    nature: str,
    tag_context: List[dict],
    dataset: List[dict],
    resources: List[Any],
    static_dataset: Optional[Any],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/asset-types"
    response = requests.post(
        url=url,
        json={
            "name": name,
            "nature": nature,
            "tagContext": tag_context,
            "dataSet": dataset,
            "resources": resources,
            "staticDataSet": static_dataset,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_asset_type(
    asset_type_id: str,
    name: str,
    nature: str,
    tag_context: List[dict],
    dataset: List[Any],
    resources: List[Any],
    static_dataset: Any,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing asset type.

    Parameters
    ----------
    asset_type_id : str
        The OIAnalytics ID of the asset type to update.
    name : str
        The name of the asset type to update.
    nature : {'DEVICE', 'PHYSICAL_ASSET', 'LOGICAL_ASSET'}
        The nature of the asset type to update.
    tag_context : list of dict
        The array of tag values used to label the asset type.
    dataset : list of dict.
        The set of metadata defined at asset type level.
    resources : list of Any.
        The set of resources defined at asset type level.
    static_dataset : list of dict
    api_credentials

    Returns
    -------
    int
        Status Code
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/asset-types/{asset_type_id}"
    response = requests.put(
        url=url,
        json={
            "name": name,
            "nature": nature,
            "tagContext": tag_context,
            "dataSet": dataset,
            "resources": resources,
            "staticDataSet": static_dataset,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def get_assets(
    asset_type_id: str,
    tag_value_id: Optional[List[str]] = None,
    query: Optional[int] = None,
    page: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "query": query,
            "assetTypeId": asset_type_id,
            "tagValueIds": tag_value_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_asset(
    asset_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets/{asset_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_single_asset_tags_and_values(
    asset_id: str,
    tag_commands: Optional[List[Dict]] = None,
    value_commands: Optional[List[Dict]] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Init commands
    if tag_commands is None:
        tag_commands = []

    if value_commands is None:
        value_commands = []

    # Build payload
    payload = [
        {
            "assetId": asset_id,
            "tagCommands": tag_commands,
            "staticDataValueCommands": value_commands,
        }
    ]

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets/tags-and-values"
    response = requests.put(url=url, json=payload, **api_credentials.auth_kwargs)

    # Output
    response.raise_for_status()
    return response.status_code


def delete_asset(
    asset_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an asset identified by its id.

    Parameters
    ----------
    asset_id : str
        The OIAnalytics ID of the asset to be deleted.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets/{asset_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def create_asset(
    name: str,
    external_reference: str,
    description: str,
    asset_type_id: str,
    tag_values: List[dict],
    data_mappings: List[dict],
    static_data_values: List[dict],
    resource_ids: List[str],
    python_model_instances: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets"
    response = requests.post(
        url=url,
        json={
            "name": name,
            "externalReference": external_reference,
            "description": description,
            "assetTypeId": asset_type_id,
            "tagValues": tag_values,
            "dataMappings": data_mappings,
            "staticDataValues": static_data_values,
            "resourceIds": resource_ids,
            "pythonModelInstances": python_model_instances,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_asset(
    asset_id: str,
    name: str,
    external_reference: str,
    description: str,
    asset_type_id: str,
    tag_values: List[dict],
    data_mappings: List[dict],
    static_data_values: List[dict],
    resource_ids: List[str],
    python_model_instances: List[dict],
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing asset.

    Parameters
    ----------
    asset_id : str
        The OIAnalytics id of the asset to update.
    name : str
        The name of the asset to create.
    external_reference : str
        The reference of the asset in external system.
    description : str
        An optional description used to understand what this asset means.
    asset_type_id : str
        The asset type this asset is attached to.
    tag_values : list of dict
        The array of tag values used to label the asset.
    data_mappings : list of dict
        The data mapping between AssetTypeData and Data for this asset.
    static_data_values : list of values
        The static data value of asset static data attached to this asset.
    resource_ids : List[str]
        The OIAnalytics ID of the resource to attach to the asset.
    python_model_instances : List[dict]
        The instances of python models to create or update.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/assets/{asset_id}"
    response = requests.put(
        url=url,
        json={
            "name": name,
            "externalReference": external_reference,
            "description": description,
            "assetTypeId": asset_type_id,
            "tagValues": tag_values,
            "dataMappings": data_mappings,
            "staticDataValues": static_data_values,
            "resourceIds": resource_ids,
            "pythonModelInstances": python_model_instances,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from typing import Optional

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = ["get_units", "get_unit_families"]


def get_units(
    label: Optional[str] = None,
    quantity_id: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_quantity: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
    List units by page.

    Parameters
    ----------
    label : str, optional
        The label the unit should contain.
    quantity_id : str, optional
        The ID of the quantity the unit is attached to.
    page : int, default 0
        The page to retrieve. It is ignored if 'get_all_pages' is set to True.
    page_size : int, default 20.
        The size of one page. It is ignored if 'get_all_pages' is set to True.
    get_all_pages : bool, default True
        Whether to get units of all pages.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_quantity: bool, default True
        Whether to expand the column 'quantity'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
        A pandas DataFrame containing the units indexed by their correspondent units IDs.

    """
    if get_all_pages:
        page = 0
        page_size = 500

    def get_page(page_num: int) -> dict:
        return endpoints.units.get_units(
            label=label,
            quantity_id=quantity_id,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict) -> pd.DataFrame:
        page_df = pd.DataFrame(page_response["content"])
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "label",
                    "quantity",
                    "multiplicativeFactor",
                    "additiveFactor",
                ]
            )
        page_df.set_index("id", inplace=True)
        return page_df

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    if expand_quantity:
        df = utils.expand_dataframe_column(
            df=df,
            col="quantity",
        )

    # output
    return df


def get_unit_families(
    label: Optional[str] = None,
    main_unit_id: Optional[str] = None,
    quantity_id: Optional[str] = None,
    page: int = 0,
    page_size: int = 20,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_main_unit: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.DataFrame:
    """
        List unit families by page.

    Parameters
    ----------
    label : str, optional
        The label the unit should contain.
    main_unit_id : str, optional
        The ID of the main unit of the family.
    quantity_id : str, optional
        The ID of the quantity the unit is attached to.
    page : int, default 0
        The page to retrieve.
    page_size : int, default 20
        The size of one page.
    get_all_pages : bool, default True
        Whether to get unit families of all pages.
        If the value is set to True, the 'page' and 'page_size' arguments will be ignored, and the function will retrieve all pages.
    multithread_pages : bool, default True
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_main_unit : bool, default True
        Whether to expand the mainUnit column.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
        A pandas DataFrame containing the unit families indexed by their correspondent unit families IDs.
    """

    if get_all_pages:
        page = 0
        page_size = 500

    def get_page(page_num: int) -> dict:
        return endpoints.units.get_unit_families(
            label=label,
            main_unit_id=main_unit_id,
            quantity_id=quantity_id,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )

    def parse_page(page_response: dict) -> pd.DataFrame:
        page_df = pd.DataFrame(page_response["content"])
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=["id", "label", "mainUnit"],
            )
        page_df.set_index("id", inplace=True)
        return page_df

    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    if expand_main_unit:
        df = utils.expand_dataframe_column(
            df=df,
            col="mainUnit",
        )

    # output
    return df



import requests
from typing import Optional, List, Any

from .. import _credentials


__all__ = [
    "get_azure_blob_sources",
    "get_single_azure_blob_source",
    "delete_azure_blob_source",
    "create_azure_blob_source",
    "update_azure_blob_source",
    "test_azure_blob_source",
]


def get_azure_blob_sources(
    page: int = 0,
    page_size: int = 20,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_single_azure_blob_source(
    azure_blob_source_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources/{azure_blob_source_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def create_azure_blob_source(
    name: str,
    tag_context: List[dict],
    enabled: bool,
    polling_rate: int,
    min_age: str,
    max_age: str,
    regex_pattern: str,
    container: str,
    account_name: str,
    custom_endpoint: Optional[Any],
    account_key: str,
    shared_access_signature: Optional[Any],
    path: str,
    preserve_files: bool,
    container_name_prepended: bool,
    include_subdirectories: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    """
    Create a new azure blob source.

    Parameters
    ----------
    name : str
        The name of the azure blob polling source.
    tag_context : List[dict]
        The tag context of the polling source.
    enabled : bool
        Whether the polling of this source is enabled.
    polling_rate : int
        The duration between two polls in seconds.
    min_age : str
        The minimum age a file should have to be considered by a poll.
    max_age : str
        The maximum age a file should have to be considered by a poll.
    regex_pattern : str
        A regex the file name should match to be considered by a poll.
    container : str
        The azure container to poll.
    account_name : str, optional
        The name of the azure account. It must be None if 'custom_endpoint' is not None.
    custom_endpoint : str, optional
        A custom endpoint to use in place of the standard azure endpoint.
        It must be None if 'account_name' is not None.
    account_key : str
        The secret account key attached to the account if the authentication mode requested is by account key.
    shared_access_signature : Any
        The shared access signature to access the resource if the authentication mode requested is by shared access signature.
    path : str
        The path to poll within the container.
    preserve_files : bool
        Whether to preserve files after a pool. If False, then files are deleted.
    container_name_prepended : bool
        If true then the name of the container is added to the file name to form the final name in OIAnalytics.
    include_subdirectories : bool
        If true look for all files also included in sub-directories of the main path that was polled.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    dict
        Details of the azure blob source that has been created.
    """

    # validation
    if account_name is None and custom_endpoint is None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be defined")
    elif account_name is not None and custom_endpoint is not None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be None")

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources"
    response = requests.post(
        url=url,
        json={
            "name": name,
            "tagContext": tag_context,
            "enabled": enabled,
            "pollingRate": polling_rate,
            "minAge": min_age,
            "maxAge": max_age,
            "regexPattern": regex_pattern,
            "container": container,
            "accountName": account_name,
            "customEndpoint": custom_endpoint,
            "accountKey": account_key,
            "sharedAccessSignature": shared_access_signature,
            "path": path,
            "preserveFiles": preserve_files,
            "containerNamePrepended": container_name_prepended,
            "includeSubdirectories": include_subdirectories,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_azure_blob_source(
    azure_blob_source_id: str,
    name: str,
    tag_context: List[dict],
    enabled: bool,
    polling_rate: int,
    min_age: str,
    max_age: str,
    regex_pattern: str,
    container: str,
    account_name: str,
    custom_endpoint: Optional[Any],
    account_key: str,
    shared_access_signature: Optional[Any],
    path: str,
    preserve_files: bool,
    container_name_prepended: bool,
    include_subdirectories: bool,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Update an existing azure blob source.

    Parameters
    ----------
    azure_blob_source_id : str
        The OIAnalytics ID of the azure blob source.
    name : str
        The name of the azure blob polling source.
    tag_context : List[dict]
        The tag context of the polling source.
    enabled : bool
        Whether the polling of this source is enabled.
    polling_rate : int
        The duration between two polls in seconds.
    min_age : str
        The minimum age a file should have to be considered by a poll.
    max_age : str
        The maximum age a file should have to be considered by a poll.
    regex_pattern : str
        A regex the file name should match to be considered by a poll.
    container : str
        The azure container to poll.
    account_name : str, optional
        The name of the azure account. It must be None if 'custom_endpoint' is not None.
    custom_endpoint : str, optional
        A custom endpoint to use in place of the standard azure endpoint.
        It must be None if 'account_name' is not None.
    account_key : str
        The secret account key attached to the account if the authentication mode requested is by account key.
    shared_access_signature : Any
        The shared access signature to access the resource if the authentication mode requested is by shared access signature.
    path : str
        The path to poll within the container.
    preserve_files : bool
        Whether to preserve files after a pool. If False, then files are deleted.
    container_name_prepended : bool
        If true then the name of the container is added to the file name to form the final name in OIAnalytics.
    include_subdirectories : bool
        If true look for all files also included in sub-directories of the main path that was polled.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """

    # validation
    if account_name is None and custom_endpoint is None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be defined")
    elif account_name is not None and custom_endpoint is not None:
        raise ValueError("one of 'account_name' and 'custom_endpoint' should be None")

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources/{azure_blob_source_id}"
    response = requests.put(
        url=url,
        json={
            "name": name,
            "tagContext": tag_context,
            "enabled": enabled,
            "pollingRate": polling_rate,
            "minAge": min_age,
            "maxAge": max_age,
            "regexPattern": regex_pattern,
            "container": container,
            "accountName": account_name,
            "customEndpoint": custom_endpoint,
            "accountKey": account_key,
            "sharedAccessSignature": shared_access_signature,
            "path": path,
            "preserveFiles": preserve_files,
            "containerNamePrepended": container_name_prepended,
            "includeSubdirectories": include_subdirectories,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def delete_azure_blob_source(
    azure_blob_source_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> int:
    """
    Delete an azure blob polling source identified by its ID.

    Parameters
    ----------
    azure_blob_source_id : str
        The OIAnalytics ID of the azure blob polling source to be deleted.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    int
        Status Code.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources/{azure_blob_source_id}"
    response = requests.delete(
        url=url,
        **api_credentials.auth_kwargs,
    )

    response.raise_for_status()
    return response.status_code


def test_azure_blob_source(
    azure_blob_source_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> dict:
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/azure-blob-polling-sources/{azure_blob_source_id}/tests"
    response = requests.post(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Optional, List

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils


__all__ = ["get_users", "get_user_details", "create_user"]


def get_users(
    page: int = 0,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_profile: bool = True,
    extract_from_access_list: Optional[str] = "synthesis",
    expand_access_list: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get configured users from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_profile: bool, default True
        Whether or not the profile information should be expanded into multiple columns.
    extract_from_access_list: {'all_access', 'tag_list', None}, default 'synthesis'
        What field should be extracted from access list information.
        'tag_list' will only extract the listed accessible tag keys.
        'all_access' will only extract whether or not the user has access to evefything.
        'synthesis' will combine the previous 2 extracting a list of accessible tag keys or 'All' for a full access.
        If None, the full dictionary is kept.
    expand_access_list: bool, default True
        Whether or not the access list information should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing users and their access information
    """

    # Args validation
    if extract_from_access_list not in ["all_access", "tag_list", "synthesis", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_access_list': {extract_from_access_list}"
        )

    if extract_from_access_list is None and expand_access_list is True:
        raise ValueError(
            "Access list cannot be expanded if 'extract_from_access_list' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0

    def get_page(page_num: int):
        page_response = endpoints.users.get_users(
            page=page_num, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])
        page_df["expirationDate"] = pd.to_datetime(
            page_df["expirationDate"], format="ISO8601"
        )
        if expand_profile is True:
            page_df = utils.expand_dataframe_column(page_df, "profile")

        # Format acess list
        if extract_from_access_list == "tag_list":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {a["tagKey"]: a["accessibleTagValues"] for a in al}
            )
        elif extract_from_access_list == "all_access":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {a["tagKey"]: a["allTagValuesAccessible"] for a in al}
            )
        elif extract_from_access_list == "synthesis":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {
                    a["tagKey"]: "All"
                    if a["allTagValuesAccessible"] is True
                    else a["accessibleTagValues"]
                    for a in al
                }
            )

        if expand_access_list is True and extract_from_access_list is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "accessList", add_prefix=False
            )

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_user_details(
    user_id: str,
    expand_profile: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Retrieve the detail of a specific user with its id.

    Parameters
    ----------
    user_id : str
        The OIAnalytics ID of the user to retrieve.
    expand_profile : bool, default True
        Whether to expand the 'profile' column.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        A pandas.Series with the details of a specific user.
    """
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    response = endpoints.users.get_user_details(
        user_id=user_id,
        api_credentials=api_credentials,
    )

    df = pd.DataFrame([response])

    if expand_profile:
        df = utils.expand_dataframe_column(
            df=df,
            col="profile",
        )

    # Output
    return df.squeeze()


def create_user(
    login: str,
    first_name: str,
    last_name: str,
    email: Optional[str],
    phone_number: Optional[str],
    language: str,
    timezone: str,
    login_allowed: bool,
    technical_user: bool,
    using_sso: bool,
    session_timeout: Optional[int],
    expiration_date: int,
    profile_id: str,
    tag_access_list: List[dict],
    expand_profile: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
) -> pd.Series:
    """
    Create a new user.

    Parameters
    ----------
    login : str
        The login of the user.
    first_name : str
        The first name of the user.
    last_name : str
        The last name of the user.
    email : str
        The email of the user.
    phone_number : str, optional
        The phone number of the user.
    language : str
        The display language of the user.
    timezone : str
        The timezone of the user.
    login_allowed : bool
        Whether the user is allowed to log into the application.
    technical_user : bool
        Whether the user is technical.
    using_sso : bool
        Whether the user is using Single Sign-On (SSO) to authenticate into the application.
    session_timeout : int, optional
        The session timeout, in minutes. Must be at least 1440.
        If None, a default timeout will be used.
    expiration_date : int
        The expiration date after which the user can no longer log in.
    profile_id : str
        The OIAnalytics ID of the user’s profile.
    tag_access_list : list of dict
        The tag access used by the user.
        It is a list of dict, where each dict contain the following keys:
        - 'tagKeyId': The tag key ID of the tag access.
        - 'allTagValuesAccessible': Whether the user has access to all the tag values of the tag key.
        - 'accessibleTagValueIds': The tag value IDs of the tag access.
    expand_profile : bool, default True
        Whether to split the 'profile' column into 'profile_in' and 'profile_name'.
    api_credentials : OIAnalyticsAPICredentials, optional
        The credentials to use to query the API.
        If None, previously set default credentials are used.

    Returns
    -------
    pd.Series
        Details of the user that has just been created.
    """

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # endpoint
    response = endpoints.users.create_user(
        login=login,
        first_name=first_name,
        last_name=last_name,
        email=email,
        phone_number=phone_number,
        language=language,
        timezone=timezone,
        login_allowed=login_allowed,
        technical_user=technical_user,
        using_sso=using_sso,
        session_timeout=session_timeout,
        expiration_date=expiration_date,
        profile_id=profile_id,
        tag_access_list=tag_access_list,
        api_credentials=api_credentials,
    )

    df = pd.DataFrame([response])

    if expand_profile:
        df = utils.expand_dataframe_column(
            df=df,
            col="profile",
        )

    return df.squeeze()
